////////////////////////////////////////////////////////////////////////////////
// TinyBasic Plus
////////////////////////////////////////////////////////////////////////////////
//
// Authors: 
//    Gordon Brandly (Tiny Basic for 68000)
//    Mike Field <hamster@snap.net.nz> (Arduino Basic) (port to Arduino)
//    Scott Lawrence <yorgle@gmail.com> (TinyBasic Plus) (features, etc)
//
// Contributors:
//          Brian O'Dell <megamemnon@megamemnon.com> (INPUT)
//    (full list tbd)

//  For full history of Tiny Basic, please see the wikipedia entry here:
//    https://en.wikipedia.org/wiki/Tiny_BASIC

// LICENSING NOTES:
//    Mike Field based his C port of Tiny Basic on the 68000 
//    Tiny BASIC which carried the following license:
/*
******************************************************************
*                                                                *
*               Tiny BASIC for the Motorola MC68000              *
*                                                                *
* Derived from Palo Alto Tiny BASIC as published in the May 1976 *
* issue of Dr. Dobb's Journal.  Adapted to the 68000 by:         *
*       Gordon Brandly                                           *
*       12147 - 51 Street                                        *
*       Edmonton AB  T5W 3G8                                     *
*       Canada                                                   *
*       (updated mailing address for 1996)                       *
*                                                                *
* This version is for MEX68KECB Educational Computer Board I/O.  *
*                                                                *
******************************************************************
*    Copyright (C) 1984 by Gordon Brandly. This program may be   *
*    freely distributed for personal use only. All commercial    *
*                      rights are reserved.                      *
******************************************************************
*/
//    ref: http://members.shaw.ca:80/gbrandly/68ktinyb.html
//
//    However, Mike did not include a license of his own for his
//    version of this.  
//    ref: http://hamsterworks.co.nz/mediawiki/index.php/Arduino_Basic
//
//    From discussions with him, I felt that the MIT license is
//    the most applicable to his intent.
//
//    I am in the process of further determining what should be
//    done wrt licensing further.  This entire header will likely
//    change with the next version 0.16, which will hopefully nail
//    down the whole thing so we can get back to implementing
//    features instead of licenses.  Thank you for your time.

#define kVersion "Ver 0.20.P"
// v0.20: 2021-04-01
//      Add exp MOD  -- a % b | 10 % 3 = 1  | 5 % 6 * 8 = 5 (Same CP/M MBASIC-85, Is not C languege)
//      Add exp AND  -- a & b | 10 & 3 = 2
//      Add Hex define $hhhh
//      Add Print Hex format  | a = 10; print $a | "A"
//                            | print $peek(x)
//      Add Function RDATE()  | return (y*256*16 + m*256 + day)
//      Add Function RTIME()  | return (hh*256*256 + mm*256 + ss)
//      Add double word access | peek(a#), poke($40001400#)
//      Add all variable data A-Z is double word
//      Add Instruction SETRTC "YYYY/MM/DD HH:MM:SS"
//      Add             SLEEP msec
//      Add Escape Sequence in Strings | \888 (Character code with octal code)
//                                     | \\ ('\'), \' or \" (Quoted String Characters)  
//      Add Limit in Input Basic Code Text | 400 charcters
//      Add command ALOAD | Rewrite Basic Statements form SD Card File (Add Statements)
//      Add Instruction ADDA A-Z, add const | Calculate for Address 
//                                          | ADDA A,1   : A <- A + 1
//                                          | ADDA A,-1  : A <- A - 1
//      Add memory Free function
//      Add External commands cp, type, edit, ....
// v0.15: 2018-06-23
//      Integrating some contributions
//      Corrected some of the #ifdef nesting atop this page
//      Licensing issues beginning to be addressed

// v0.14: 2013-11-07
//      Modified Input command to accept an expression using getn()
//      Syntax is "input x" where x is any variable
//      NOTE: This only works for numbers, expressions. not strings.
//
// v0.13: 2013-03-04
//      Support for Arduino 1.5 (SPI.h included, additional changes for DUE support)
//
// v0.12: 2013-03-01
//      EEPROM load and save routines added: EFORMAT, ELIST, ELOAD, ESAVE, ECHAIN
//      added EAUTORUN option (chains to EEProm saved program on startup)
//      Bugfixes to build properly on non-arduino systems (PROGMEM #define workaround)
//      cleaned up a bit of the #define options wrt TONE
//
// v0.11: 2013-02-20
//      all display strings and tables moved to PROGMEM to save space
//      removed second serial
//      removed pinMode completely, autoconf is explicit
//      beginnings of EEPROM related functionality (new,load,save,list)
//
// v0.10: 2012-10-15
//      added kAutoConf, which eliminates the "PINMODE" statement.
//      now, DWRITE,DREAD,AWRITE,AREAD automatically set the PINMODE appropriately themselves.
//      should save a few bytes in your programs.
//
// v0.09: 2012-10-12
//      Fixed directory listings.  FILES now always works. (bug in the SD library)
//      ref: http://arduino.cc/forum/index.php/topic,124739.0.html
//      fixed filesize printouts (added printUnum for unsigned numbers)
//      #defineable baud rate for slow connection throttling
//e
// v0.08: 2012-10-02
//      Tone generation through piezo added (TONE, TONEW, NOTONE)
//
// v0.07: 2012-09-30
//      Autorun buildtime configuration feature
//
// v0.06: 2012-09-27
//      Added optional second serial input, used for an external keyboard
//
// v0.05: 2012-09-21
//      CHAIN to load and run a second file
//      RND,RSEED for random stuff
//      Added "!=" for "<>" synonym
//      Added "END" for "STOP" synonym (proper name for the functionality anyway)
//
// v0.04: 2012-09-20
//      DELAY ms   - for delaying
//      PINMODE <pin>, INPUT|IN|I|OUTPUT|OUT|O
//      DWRITE <pin>, HIGH|HI|1|LOW|LO|0
//      AWRITE <pin>, [0..255]
//      fixed "save" appending to existing files instead of overwriting
// 	Updated for building desktop command line app (incomplete)
//
// v0.03: 2012-09-19
//	Integrated Jurg Wullschleger whitespace,unary fix
//	Now available through github
//	Project renamed from "Tiny Basic in C" to "TinyBasic Plus"
//	   
// v0.02b: 2012-09-17  Scott Lawrence <yorgle@gmail.com>
// 	Better FILES listings
//
// v0.02a: 2012-09-17  Scott Lawrence <yorgle@gmail.com>
// 	Support for SD Library
// 	Added: SAVE, FILES (mostly works), LOAD (mostly works) (redirects IO)
// 	Added: MEM, ? (PRINT)
// 	Quirk:  "10 LET A=B+C" is ok "10 LET A = B + C" is not.
// 	Quirk:  INPUT seems broken?

// IF testing with Visual C, this needs to be the first thing in the file.
//#include "stdafx.h"

/* Evauated Board configurations */
#include "ff.h"
#include "Ev_Board_conf.h"

char eliminateCompileErrors = 1;  // fix to suppress arduino build errors

// hack to let makefiles work with this file unchanged
#ifdef RASPI_PICO
#undef ARDUINO
#else

#ifdef FORCE_DESKTOP
#undef ARDUINO
#else
#define ARDUINO 1
#endif

#endif

////////////////////////////////////////////////////////////////////////////////
// Feature option configuration...

// This enables LOAD, SAVE, FILES commands through the Arduino SD Library
// it adds 9k of usage as well.
#define ENABLE_FILEIO 1
//#undef ENABLE_FILEIO

// this turns on "autorun".  if there's FileIO, and a file "autorun.bas",
// then it will load it and run it when starting up
//#define ENABLE_AUTORUN 1
#undef ENABLE_AUTORUN
// and this is the file that gets run
#define kAutorunFilename  "autorun.bas"

// this is the alternate autorun.  Autorun the program in the eeprom.
// it will load whatever is in the EEProm and run it
#define ENABLE_EAUTORUN 1
//#undef ENABLE_EAUTORUN

// this will enable the "TONE", "NOTONE" command using a piezo
// element on the specified pin.  Wire the red/positive/piezo to the kPiezoPin,
// and the black/negative/metal disc to ground.
// it adds 1.5k of usage as well.
//#define ENABLE_TONES 1
#undef ENABLE_TONES
#define kPiezoPin 5

// we can use the EEProm to store a program during powerdown.  This is 
// 1kbyte on the '328, and 512 bytes on the '168.  Enabling this here will
// allow for this funcitonality to work.  Note that this only works on AVR
// arduino.  Disable it for DUE/other devices.
#define ENABLE_EEPROM 1
//#undef ENABLE_EEPROM

// Sometimes, we connect with a slower device as the console.
// Set your console D0/D1 baud rate here (9600 baud default)
#define kConsoleBaud 9600

#define GETLN_MAX 400

////////////////////////////////////////////////////////////////////////////////
// fixes for RAMEND on some platforms
#ifndef RAMEND
  // RAMEND is defined for Uno type Arduinos
  #ifdef ARDUINO
    // probably DUE or 8266?
    #ifdef ESP8266
      #define RAMEND (8192-1)
    #else
      // probably DUE - ARM rather than AVR
      #define RAMEND (4096-1)
    #endif
  #endif
#endif


// Enable memory alignment for certain processers (e.g. some ESP8266-based devices)
#if  defined(ESP8266) || defined(RASPI_PICO)
  // Uses up to one extra byte per program line of memory
  #define ALIGN_MEMORY 1
#else
  #undef ALIGN_MEMORY
#endif

#ifndef ARDUINO
  // not an arduino, so we can disable these features.
  // turn off EEProm
  #undef ENABLE_EEPROM
  #undef ENABLE_TONES
#endif


// includes, and settings for Arduino-specific features
#ifdef ARDUINO

  // EEPROM
  #ifdef ENABLE_EEPROM
    #include <EEPROM.h>  /* NOTE: case sensitive */
    int eepos = 0;
  #endif

  // SD card File io
  #ifdef ENABLE_FILEIO
    #include <SD.h>
    #include <SPI.h> /* needed as of 1.5 beta */

    // set this to the card select for your Arduino SD shield
    #define kSD_CS 10

    #define kSD_Fail  0
    #define kSD_OK    1

    File fp;
  #endif

  // set up our RAM buffer size for program and user input
  // NOTE: This number will have to change if you include other libraries.
  //       It is also an estimation.  Might require adjustments...
  #ifdef ENABLE_FILEIO
    #define kRamFileIO (1030) /* approximate */
  #else
    #define kRamFileIO (0)
  #endif

  #ifdef ENABLE_TONES
    #define kRamTones (40)
  #else
    #define kRamTones (0)
  #endif

  #define kRamSize  (RAMEND - 1160 - kRamFileIO - kRamTones) 

#endif /* ARDUINO Specifics */


// set up file includes for things we need, or desktop specific stuff.

#ifdef ARDUINO
  // Use pgmspace/PROGMEM directive to store strings in progmem to save RAM
  #include <avr/pgmspace.h>
#else
  #include <stdio.h>
  #include <stdlib.h>

  #undef ENABLE_TONES

  // size of our program ram
  #define kRamSize   64*1024 /* arbitrary - not dependant on libraries */

#ifndef RASPI_PICO
  #ifdef ENABLE_FILEIO
    FILE * fp;
  #endif
#endif
#endif

////////////////////

// memory alignment
//  necessary for some esp8266-based devices
#ifdef ALIGN_MEMORY
  // Align memory addess x to an even page
  #define ALIGN_UP(x) ((unsigned char*)(((unsigned int)(x + 1) >> 1) << 1))
  #define ALIGN_DOWN(x) ((unsigned char*)(((unsigned int)x >> 1) << 1))
#else
  #define ALIGN_UP(x) x
  #define ALIGN_DOWN(x) x
#endif


////////////////////
// various other desktop-tweaks and such.

#ifndef boolean 
  #define boolean int
  #define true 1
  #define false 0
#endif

#ifndef byte
  typedef unsigned char byte;
#endif

// some catches for AVR based text string stuff...
#ifndef PROGMEM
  #define PROGMEM
#endif
#ifndef pgm_read_byte
  #define pgm_read_byte( A ) *(A)
#endif

////////////////////

#ifdef ENABLE_FILEIO
#ifdef RASPI_PICO
  void cmd_Files( char *dir );
#else
  // functions defined elsehwere
  void cmd_Files( void );
  static boolean sd_is_initialized = false;
#endif
  void filenameWord(unsigned char *name);
#endif

// some settings based things

boolean inhibitOutput = false;
static boolean runAfterLoad = false;
static boolean triggerRun = false;

// these will select, at runtime, where IO happens through for load/save
enum {
  kStreamSerial = 0,
  kStreamEEProm,
  kStreamFile
};
static unsigned char inStream = kStreamSerial;
static unsigned char outStream = kStreamSerial;


////////////////////////////////////////////////////////////////////////////////
// ASCII Characters
#define CR	'\r'
#define NL	'\n'
#define LF      0x0a
#define TAB	'\t'
#define BELL	'\b'
#define SPACE   ' '
#define SQUOTE  '\''
#define DQUOTE  '\"'
#define CTRLC	0x03
#define CTRLH	0x08
#define CTRLS	0x13
#define CTRLX	0x18
#ifdef RASPI_PICO
#define CTRLBS	0x7F
#define ESCHEX  '$'
#endif

typedef short unsigned LINENUM;
#ifdef ARDUINO
#define ECHO_CHARS 1
#else
#define ECHO_CHARS 0
#endif


static unsigned char program[kRamSize];
static const char *  sentinel = "HELLO";
static unsigned char *txtpos,*list_line, *tmptxtpos;
static unsigned char expression_error;
static unsigned char *tempsp;

/***********************************************************/
// Keyword table and constants - the last character has 0x80 added to it
const static unsigned char keywords[] PROGMEM = {
  'L','I','S','T'+0x80,
  'L','O','A','D'+0x80,
  'N','E','W'+0x80,
  'R','U','N'+0x80,
  'S','A','V','E'+0x80,
  'N','E','X','T'+0x80,
  'L','E','T'+0x80,
  'I','F'+0x80,
  'G','O','T','O'+0x80,
  'G','O','S','U','B'+0x80,
  'R','E','T','U','R','N'+0x80,
  'R','E','M'+0x80,
  'F','O','R'+0x80,
  'I','N','P','U','T'+0x80,
  'P','R','I','N','T'+0x80,
  'P','O','K','E'+0x80,
  'S','T','O','P'+0x80,
  'B','Y','E'+0x80,
  'F','I','L','E','S'+0x80,
  'M','E','M'+0x80,
  '?'+ 0x80,
  '\''+ 0x80,
#ifndef RASPI_PICO
  'A','W','R','I','T','E'+0x80,
  'D','W','R','I','T','E'+0x80,
  'D','E','L','A','Y'+0x80,
#endif
  'E','N','D'+0x80,
  'R','S','E','E','D'+0x80,
#ifndef RASPI_PICO
  'C','H','A','I','N'+0x80,
#else // Yes RASPI_PICO
  'P','I','N','M','O','D','E'+0x80,
  'S','E','T','R','T','C'+0x80,
  'S','L','E','E','P'+0x80,
  'A','D','D','A'+0x80,
  'A','L','O','A','D'+0x80,
  '!' + 0x80,
#endif
#ifdef ENABLE_TONES
  'T','O','N','E','W'+0x80,
  'T','O','N','E'+0x80,
  'N','O','T','O','N','E'+0x80,
#endif
#ifdef ARDUINO
#ifdef ENABLE_EEPROM
  'E','C','H','A','I','N'+0x80,
  'E','L','I','S','T'+0x80,
  'E','L','O','A','D'+0x80,
  'E','F','O','R','M','A','T'+0x80,
  'E','S','A','V','E'+0x80,
#endif
#endif
  0
};

// by moving the command list to an enum, we can easily remove sections 
// above and below simultaneously to selectively obliterate functionality.
enum {
  KW_LIST = 0,
  KW_LOAD, KW_NEW, KW_RUN, KW_SAVE,
  KW_NEXT, KW_LET, KW_IF,
  KW_GOTO, KW_GOSUB, KW_RETURN,
  KW_REM,
  KW_FOR,
  KW_INPUT, KW_PRINT,
  KW_POKE,
  KW_STOP, KW_BYE,
  KW_FILES,
  KW_MEM,
  KW_QMARK, KW_QUOTE,
#ifndef RASPI_PICO
  KW_AWRITE, KW_DWRITE,
  KW_DELAY,
#endif
  KW_END,
  KW_RSEED,
#ifndef RASPI_PICO
  KW_CHAIN,
#else // Yes RASPI_PICO
  KW_PINMODE,
  KW_SETRTC,
  KW_SLEEP,
  KW_ADDA,
  KW_ALOAD,
  KW_FILECMD,
#endif
#ifdef ENABLE_TONES
  KW_TONEW, KW_TONE, KW_NOTONE,
#endif
#ifdef ARDUINO
#ifdef ENABLE_EEPROM
  KW_ECHAIN, KW_ELIST, KW_ELOAD, KW_EFORMAT, KW_ESAVE, 
#endif
#endif
  KW_DEFAULT /* always the final one*/
};

struct stack_for_frame {
  char frame_type;
  char for_var;
  long terminal;
  long step;
  unsigned char *current_line;
  unsigned char *txtpos;
};

struct stack_gosub_frame {
  char frame_type;
  unsigned char *current_line;
  unsigned char *txtpos;
};

const static unsigned char func_tab[] PROGMEM = {
  'P','E','E','K'+0x80,
  'A','B','S'+0x80,
  'A','R','E','A','D'+0x80,
  'D','R','E','A','D'+0x80,
  'R','N','D'+0x80,
  'R','D','A','T','E'+0x80,
  'R','T','I','M','E'+0x80,
   0
};
#define FUNC_PEEK    0
#define FUNC_ABS     1
#define FUNC_AREAD   2
#define FUNC_DREAD   3
#define FUNC_RND     4
#define FUNC_RDATE   5
// RDATE() -> yyyy*256*16+m*256+d
#define FUNC_RTIME   6
// RDATE() -> hour*256*256+min*256+ss
#define FUNC_UNKNOWN 7

const static unsigned char to_tab[] PROGMEM = {
  'T','O'+0x80,
  0
};

const static unsigned char step_tab[] PROGMEM = {
  'S','T','E','P'+0x80,
  0
};

const static unsigned char relop_tab[] PROGMEM = {
  '>','='+0x80,
  '<','>'+0x80,
  '>'+0x80,
  '='+0x80,
  '<','='+0x80,
  '<'+0x80,
  '!','='+0x80,
  0
};

#define RELOP_GE		0
#define RELOP_NE		1
#define RELOP_GT		2
#define RELOP_EQ		3
#define RELOP_LE		4
#define RELOP_LT		5
#define RELOP_NE_BANG		6
#define RELOP_UNKNOWN	7

const static unsigned char highlow_tab[] PROGMEM = { 
  'H','I','G','H'+0x80,
  'H','I'+0x80,
  '1'+0x80,
  'L','O','W'+0x80,
  'L','O'+0x80,
  '0'+0x80,
  0
};
#define HIGHLOW_HIGH    2
#define HIGHLOW_UNKNOWN 6

const static unsigned char inout_tab[] PROGMEM = { 
  'O','U','T'+0x80,
  '1'+0x80,
  'I','N'+0x80,
  '0'+0x80,
  0
};
#define INOUT_OUT    1
#define INOUT_UNKNOWN 4


#define STACK_SIZE (sizeof(struct stack_for_frame)*9)  // Stack Size is 9 
#define VAR_SIZE sizeof(long) // Size of variables in bytes

static unsigned char *stack_limit;
static unsigned char *program_start;
static unsigned char *program_end;
static unsigned char *stack; // Software stack for things that should go on the CPU stack
static unsigned char *variables_begin;
static unsigned char *current_line;
static unsigned char *sp;
#define STACK_GOSUB_FLAG 'G'
#define STACK_FOR_FLAG 'F'
static unsigned char table_index;
static LINENUM linenum;
static unsigned char *next_var = NULL;
#ifdef RASPI_PICO
static unsigned char basic_cmd;
#endif

static const unsigned char okmsg[]            PROGMEM = "OK";
static const unsigned char whatmsg[]          PROGMEM = "What? ";
static const unsigned char howmsg[]           PROGMEM =	"How?";
static const unsigned char sorrymsg[]         PROGMEM = "Sorry!";
static const unsigned char initmsg[]          PROGMEM = "TinyBasic Plus " kVersion;
static const unsigned char memorymsg[]        PROGMEM = " bytes free.";
#ifdef ARDUINO
#ifdef ENABLE_EEPROM
static const unsigned char eeprommsg[]        PROGMEM = " EEProm bytes total.";
static const unsigned char eepromamsg[]       PROGMEM = " EEProm bytes available.";
#endif
#endif
static const unsigned char breakmsg[]         PROGMEM = "break!";
static const unsigned char unimplimentedmsg[] PROGMEM = "Unimplemented";
static const unsigned char backspacemsg[]     PROGMEM = "\b \b";
static const unsigned char indentmsg[]        PROGMEM = "    ";
static const unsigned char sderrormsg[]       PROGMEM = "SD card error.";
static const unsigned char sdfilemsg[]        PROGMEM = "SD file error.";
static const unsigned char dirextmsg[]        PROGMEM = "(dir)";
static const unsigned char slashmsg[]         PROGMEM = "/";
static const unsigned char spacemsg[]         PROGMEM = " ";

static int inchar(void);
static void outchar(unsigned char c);
static void line_terminator(void);
static long expression(void);
static unsigned char breakcheck(void);
/***************************************************************************/
static void ignore_blanks(void)
{
  while(*txtpos == SPACE || *txtpos == TAB)
    txtpos++;
}


/***************************************************************************/
static void scantable(const unsigned char *table)
{
  int i = 0;
  table_index = 0;
  while(1)
  {
    // Run out of table entries?
    if(pgm_read_byte( table ) == 0)
      return;

    // Do we match this character?
    if(txtpos[i] == pgm_read_byte( table ))
    {
      i++;
      table++;
    }
    else
    {
      // do we match the last character of keywork (with 0x80 added)? If so, return
      if(txtpos[i]+0x80 == pgm_read_byte( table ))
      {
        txtpos += i+1;  // Advance the pointer to following the keyword
        ignore_blanks();
        return;
      }

      // Forward to the end of this keyword
      while((pgm_read_byte( table ) & 0x80) == 0)
        table++;

      // Now move on to the first character of the next word, and reset the position index
      table++;
      table_index++;
      ignore_blanks();
      i = 0;
    }
  }
}

/***************************************************************************/
static void pushb(unsigned char b)
{
  sp--;
  *sp = b;
}

/***************************************************************************/
static unsigned char popb()
{
  unsigned char b;
  b = *sp;
  sp++;
  return b;
}

/***************************************************************************/
void printnum(int num, int d_min)
{
  int digits = 0;

  if(num < 0)
  {
    num = -num;
    outchar('-');
  }
  do {
    pushb(num%10+'0');
    num = num/10;
    digits++;
  }
  while (num > 0);
  while ( digits < d_min ) { outchar('0'); d_min--; }
  while(digits > 0)
  {
    outchar(popb());
    digits--;
  }
}
void printhex(int num, int d_min)
{
  int digits = 0;
  unsigned int w = (unsigned int)num;
  do {
    char c = w%16;
    if ( c > 9 )  c += 0x37;
    else          c += 0x30;
    pushb(c);
    w = w/16;
    digits++;
  }
  while (w > 0);
  
  outchar('h');
  outchar('\'');
  while ( digits < d_min ) { outchar('0'); d_min--; }
  while(digits > 0)
  {
    outchar(popb());
    digits--;
  }
}

void printUnum(unsigned int num)
{
  int digits = 0;

  do {
    pushb(num%10+'0');
    num = num/10;
    digits++;
  }
  while (num > 0);

  while(digits > 0)
  {
    outchar(popb());
    digits--;
  }
}

/***************************************************************************/
static unsigned long testnum(void)
{ // line no. 1~65533 = OK
  // line no. 65534 = No Number
  // line no. 65535 = ERROR 0, >= 65534  
  unsigned long num = 0;
  char  f = 0;
  ignore_blanks();

  while(*txtpos>= '0' && *txtpos <= '9' )
  {
    f++;
    num = num *10 + *txtpos - '0';
    txtpos++;
    // Trap overflows
    if(num >= 0xFFFE)
    {
      num = 0xFFFF;
      DBG_NOP();
      break;
    }
  }
  if( f )  {
    if( num ) return	num;
    else      return 0xFFFF;
  } 
  else      return 0xFFFE;
}

/***************************************************************************/
static unsigned char print_quoted_string(void)
{
  int i=0;
  unsigned char c;

  unsigned char delim = *txtpos;
  if(delim != '"' && delim != '\'')
    return 0;
  txtpos++;

  // Check we have a closing delimiter
  while(txtpos[i] != delim)
  {
    if(txtpos[i] == '\\') {
       if(txtpos[i+1] == '"' || txtpos[i+1] == '\'' ) {i++;}
       else if(txtpos[i+1] == '\\' ) {i++;}
       else if((txtpos[i+1] >= '0' && txtpos[i+1] <= '3') &&
          (txtpos[i+2] >= '0' && txtpos[i+2] <= '7') &&
          (txtpos[i+3] >= '0' && txtpos[i+3] <= '7') ) {
           i += 3;
       } else {
         return 0;
       }
    }
    if(txtpos[i] == NL)
      return 0;
    i++;
  }

  // Print the characters
  while(*txtpos != delim)
  {
    if(*txtpos == '\\') {
       if( txtpos[1] == '"' || txtpos[1] == '\'' ) {txtpos++;}
       else if(txtpos[1] == '\\' ) {txtpos++;}
       else if((txtpos[1] >= '0' && txtpos[1] <= '3') &&
          (txtpos[2] >= '0' && txtpos[2] <= '7') &&
          (txtpos[3] >= '0' && txtpos[3] <= '7') ) {
           c = (txtpos[3] - '0') | ((txtpos[2] - '0')<<3) |((txtpos[1] - '0')<<6);
           outchar(c);
           txtpos += 4;
           continue;
       } else {
         return 0;
       }
    }
    outchar(*txtpos);
    txtpos++;
  }
  txtpos++; // Skip over the last delimiter

  return 1;
}

/***************************************************************************/
static unsigned char get_quoted_string(unsigned char *str, int siz)
{
  int i=0;
  unsigned char delim = *txtpos;
  if(delim != '"' && delim != '\'')
    return 0;
  txtpos++;

  // Check we have a closing delimiter
  while(txtpos[i] != delim)
  {
    if(txtpos[i] == NL)
      return 0;
    i++;
  }

  // get the characters
  i = 0;
  while(*txtpos != delim)
  {
    *str++ = *txtpos;
    txtpos++;
    i++;
    if( i >= siz) return 0; 
  }
  *str = 0;
  txtpos++; // Skip over the last delimiter

  return 1;
}

/***************************************************************************/
void printmsgNoNL(const unsigned char *msg)
{
  while( pgm_read_byte( msg ) != 0 ) {
    outchar( pgm_read_byte( msg++ ) );
  };
}

/***************************************************************************/
void printmsg(const unsigned char *msg)
{
  printmsgNoNL(msg);
  line_terminator();
}

/***************************************************************************/
static int getln(char prompt)
{
  int l = 0;

  outchar(prompt);
  txtpos = program_end;
  while(1)
  {
    unsigned char c = inchar();
    switch(c)
    {
    case NL:
      //break;
    case CR:
      while (txtpos != program_end ) {
          if ( *(txtpos-1) == ' ' )  { txtpos--; l--; }
          else                       break;
      }
      line_terminator();
      // Terminate all strings with a NL
      txtpos[0] = NL;
      return (l+1);
    case CTRLH:
    case CTRLBS:
      if(txtpos == program_end )
        break;
      txtpos--; l--;

      printmsgNoNL(backspacemsg);
      break;
    default:
      if( (c >= 0x20 && c < 0x7f) || (c >= 0x80 && c < 0xff) ) { 
      // We need to leave at least one space to allow us to shuffle the line into order
        if(txtpos == variables_begin-2)
          outchar(BELL);
        else
        {
          if( l < GETLN_MAX ) {
            txtpos[0] = c;
            txtpos++; l++;
            outchar(c);
          }
        }
      }
    }
  }
}

/***************************************************************************/
static unsigned char *findline(void)
{
  unsigned char *line = program_start;
  while(1)
  {
    if(line == program_end)
      return line;

    if(((LINENUM *)line)[0] >= linenum)
      return line;

    // Add the line length onto the current address, to get to the next line;
    line += *((unsigned short*)(line+sizeof(LINENUM)));
  }
}
/***************************************************************************/
static unsigned char *findNumberline(void)
{
  unsigned char *line = program_start;
  while(1)
  {
    if(line == program_end)
      return NULL;

    if(((LINENUM *)line)[0] == linenum)
      return line;
    if(((LINENUM *)line)[0] > linenum)
      return NULL;

    // Add the line length onto the current address, to get to the next line;
    line += *((unsigned short*)(line+sizeof(LINENUM)));
  }
}

/***************************************************************************/
static void toUppercaseBuffer(void)
{
  unsigned char *c = program_end;
  unsigned char quote = 0;

  while(*c != NL)
  {
    // Are we in a quoted string?
    if(*c == quote)
      quote = 0;
    else if(*c == '"' || *c == '\'')
      quote = *c;
    else if( quote && *c == '\\' ) {
      if( *(c+1) == '\\' || *(c+1) == '"' || *(c+1) == '\'') c++;
    }
    else if(quote == 0 && *c >= 'a' && *c <= 'z')
      *c = *c + 'A' - 'a';
    c++;
  }
}

/***************************************************************************/
void printline()
{
  LINENUM line_num = *((LINENUM *)(list_line));
  unsigned short pln = *((unsigned short *)(list_line+sizeof(LINENUM)));
  unsigned char *nxt_line = list_line + pln;

  list_line += sizeof(LINENUM) + sizeof(unsigned short);

  // Output the line */
  printnum(line_num, -1);
  outchar(' ');
  while(*list_line != NL)
  {
    outchar(*list_line);
    list_line++;
  }
  list_line++;
#ifdef ALIGN_MEMORY
  // Start looking for next line on even page
  //if (ALIGN_UP(list_line) != list_line)
  //  list_line++;
#endif
  list_line = nxt_line;
  line_terminator();
}

/***************************************************************************/
static long expr4(void)
{
  // fix provided by Jurg Wullschleger wullschleger@gmail.com
  // fixes whitespace and unary operations
  long a = 0;

  ignore_blanks();

  if( *txtpos == '-' ) {
    txtpos++;
    return -expr4();
  }
  // end fix

  //if(*txtpos == '0')
  //{
  //  return 0;
  //}
  if(*txtpos == ESCHEX) {
    txtpos++;
    while((*txtpos >= '0' && *txtpos <= '9') ||(*txtpos >= 'A' && *txtpos <= 'F') )
    {
        a = (a*16) + (*txtpos<'A'?(*txtpos - '0'):(*txtpos - 0x37));
        txtpos++;
    } 
    return a;
  }
  if(*txtpos >= '0' && *txtpos <= '9')
  {
    do 	{
      a = a*10 + *txtpos - '0';
      txtpos++;
    } 
    while(*txtpos >= '0' && *txtpos <= '9');
    return a;
  }

  // Is it a function or variable reference?
  if(txtpos[0] >= 'A' && txtpos[0] <= 'Z')
  {
    // Is it a variable reference (single alpha)
    if(txtpos[1] < 'A' || txtpos[1] > 'Z')
    {
      a = ((long *)variables_begin)[*txtpos - 'A'];
      txtpos++;
      return a;
    }

    // Is it a function with a single parameter
    scantable(func_tab);
    if(table_index == FUNC_UNKNOWN)
      goto expr4_error;

    unsigned char f = table_index;
    boolean          l = false;

    if(*txtpos != '(')
      goto expr4_error;

    txtpos++;
    if( f==FUNC_RDATE || f==FUNC_RTIME ) {
      if(*txtpos != ')') 
        goto expr4_error;
    } else {
      a = expression();
      if(*txtpos == '#') {
        l = true;
        txtpos++;
      }
      if(*txtpos != ')') {
        goto expr4_error;
      }
    }
    txtpos++;
    switch(f)
    {
    case FUNC_PEEK:
#ifdef RASPI_PICO
      if( l ) {
        if((a & 0x00000003))
          goto  expr4_error;
        return *((unsigned long*)a);
      }
      else
        return *((unsigned char*)a);
#else
      return program[a];
#endif
      
    case FUNC_ABS:
      if(a < 0) 
        return -a;
      return a;

#ifdef RASPI_PICO
    //case FUNC_AREAD:
    //  pinMode( a, INPUT );
    //  return analogRead( a );                        
    case FUNC_DREAD:
      return gpio_get(a);
    case FUNC_RDATE:
      return get_rtc_date();
    case FUNC_RTIME:
      return get_rtc_time();

#endif
#ifdef ARDUINO
    case FUNC_AREAD:
      pinMode( a, INPUT );
      return analogRead( a );                        
    case FUNC_DREAD:
      pinMode( a, INPUT );
      return digitalRead( a );
#endif

    case FUNC_RND:
#ifdef ARDUINO
      return( random( a ));
#else
      return( rand() % a );
#endif
    }
  }

  if(*txtpos == '(')
  {
    long a;
    txtpos++;
    a = expression();
    if(*txtpos != ')')
      goto expr4_error;

    txtpos++;
    return a;
  }

expr4_error:
  expression_error = 1;
  return 0;

}

/***************************************************************************/
static long expr3(void)
{
  long a,b;

  a = expr4();

  ignore_blanks(); // fix for eg:  100 a = a + 1

  while(1)
  {
    if(*txtpos == '*')
    {
      txtpos++;
      b = expr4();
      a *= b;
    }
    else if(*txtpos == '/')
    {
      txtpos++;
      b = expr4();
      if(b != 0)
        a /= b;
      else
        expression_error = 1;
    }
    else
      return a;
  }
}

/***************************************************************************/
static long expr2(void)
{
  long a,b;

  if(*txtpos == '-' || *txtpos == '+' || *txtpos == '&' || *txtpos =='%' )
    a = 0;
  else
    a = expr3();

  while(1)
  {
    if(*txtpos == '-')
    {
      txtpos++;
      b = expr3();
      a -= b;
    }
    else if(*txtpos == '+')
    {
      txtpos++;
      b = expr3();
      a += b;
    }
    else if(*txtpos == '&')
    {
      txtpos++;
      b = expr3();
      a &= b;
    }
    else if(*txtpos == '%')
    {
      txtpos++;
      b = expr3();
      a %= b;
    }
    else
      return a;
  }
}
/***************************************************************************/
static long expression(void)
{
  long a,b;

  a = expr2();

  // Check if we have an error
  if(expression_error)	return a;

  scantable(relop_tab);
  if(table_index == RELOP_UNKNOWN)
    return a;

  switch(table_index)
  {
  case RELOP_GE:
    b = expr2();
    if(a >= b) return 1;
    break;
  case RELOP_NE:
  case RELOP_NE_BANG:
    b = expr2();
    if(a != b) return 1;
    break;
  case RELOP_GT:
    b = expr2();
    if(a > b) return 1;
    break;
  case RELOP_EQ:
    b = expr2();
    if(a == b) return 1;
    break;
  case RELOP_LE:
    b = expr2();
    if(a <= b) return 1;
    break;
  case RELOP_LT:
    b = expr2();
    if(a < b) return 1;
    break;
  }
  return 0;
}
static void init_var(void)
{
  for( char v = 'A'; v < 'Z'; v++ ) {
    ((long*)variables_begin)[v-'A'] = 0;
  }
}
/***************************************************************************/
void loop()
{
  unsigned char *start;
  unsigned char *newEnd;
  unsigned short linelen;
  boolean isDigital;
  boolean alsoWait = false;
  long val;

#ifdef ARDUINO
#ifdef ENABLE_TONES
  noTone( kPiezoPin );
#endif
#endif

  program_start = program;
  program_end = program_start;
  init_var();
  sp = program+sizeof(program);  // Needed for printnum
#ifdef ALIGN_MEMORY
  // Ensure these memory blocks start on even pages
  stack_limit = ALIGN_DOWN(program+sizeof(program)-STACK_SIZE);
  variables_begin = ALIGN_DOWN(stack_limit - 27*VAR_SIZE);
#else
  stack_limit = program+sizeof(program)-STACK_SIZE;
  variables_begin = stack_limit - 27*VAR_SIZE;
#endif
  // memory free
  printnum(variables_begin-program_end, -1);
  printmsg(memorymsg);
#ifdef ARDUINO
#ifdef ENABLE_EEPROM
  // eprom size
  printnum( E2END+1, -1 );
  printmsg( eeprommsg );
#endif /* ENABLE_EEPROM */
#endif /* ARDUINO */

warmstart:
  // this signifies that it is running in 'direct' mode.
  current_line = 0;
  sp = program+sizeof(program);
  printmsg(okmsg);

prompt:
  if( triggerRun ){
    triggerRun = false;
    current_line = program_start;
    goto execline;
  }

  linelen = getln( '>' );
  toUppercaseBuffer();

  txtpos = program_end;

  // Move it to the end of program_memory
  {
    unsigned char *dest;
    unsigned char *ssrc;
    unsigned int  len = linelen;

    dest = variables_begin - 1;
    ssrc = txtpos + linelen - 1;
    while(len--)
    {
      *dest-- = *ssrc--;
    }
    txtpos = dest + 1;
  }
  // program_end ->|[BASIC TEXT] ===MOVE===> [BASIC TEXT]|<- variables_begin 

  // Now see if we have a line number
  linenum = testnum();
  ignore_blanks();
  if(linenum == 0xfffe)
    goto direct;

  if(linenum == 0xFFFF)
    goto qhow;

  // Find the length of what is left, including the (yet-to-be-populated) line header
  linelen = (variables_begin - txtpos) ;

  // Line starts should always be on 16-bit pages
  // Ajust Odd Address & length
  {
    // As the start of the line has moved, the data should move as well
    int dx = ((unsigned long)txtpos & 1);
    unsigned char *tomove = txtpos - dx;
    unsigned char *stmnt = tomove;
    int            s = linelen;
    linelen += dx;

    while ( dx != 0 && s--)
    {
      *tomove++ = *txtpos++;
    }
    txtpos = stmnt - sizeof(unsigned short) - sizeof(unsigned short);
  }
  linelen += sizeof(unsigned short) + sizeof(unsigned short); // Add space for the line no. and line length
  *((unsigned short *)txtpos) = linenum;
  *((unsigned short *)(txtpos+sizeof(LINENUM))) = linelen;
  // [LINE NO.][LINE LENGTH][BASIC TEXT][+ajust byte]|<-Memory End

  // Merge it into the rest of the program
  start = findline();

  // If a line with that number exists, then remove it
  if(start != program_end && *((LINENUM *)start) == linenum)
  {
    unsigned char *dest, *from;
    unsigned tomove;

    from = start + *((unsigned short*)(start+sizeof(LINENUM)));
    dest = start;

    tomove = program_end - from;
    while( tomove > 0)
    {
      *dest = *from;
      from++;
      dest++;
      tomove--;
    }	
    program_end = dest;
  }

  if(txtpos[sizeof(LINENUM)+sizeof(unsigned short)] == NL) // If the line has no txt, it was just a delete
    goto prompt;

  // Make room for the new line, either all in one hit or lots of little shuffles
  while(linelen > 0)
  {	
    unsigned int tomove;
    unsigned char *from,*dest;
    unsigned int space_to_make;

    space_to_make = txtpos - program_end;

    if(space_to_make > linelen)
      space_to_make = linelen;
    newEnd = program_end+space_to_make;
    tomove = program_end - start;


    // Source and destination - as these areas may overlap we need to move bottom up
    from = program_end;
    dest = newEnd;
    while(tomove > 0)
    {
      from--;
      dest--;
      *dest = *from;
      tomove--;
    }

    // Copy over the bytes into the new space
    for(tomove = 0; tomove < space_to_make; tomove++)
    {
      *start = *txtpos;
      txtpos++;
      start++;
      linelen--;
    }
    program_end = newEnd;
  }
  goto prompt;

unimplemented:
  printmsg(unimplimentedmsg);
  goto prompt;

qhow:	
  printmsg(howmsg);
  goto prompt;

qwhat:	
  printmsgNoNL(whatmsg);
  if(current_line != NULL)
  {
    unsigned char tmp = *txtpos;
    if(*txtpos != NL)
      *txtpos = '^';
    list_line = current_line;
    printline();
    *txtpos = tmp;
  } else {
    line_terminator();
  }
  goto prompt;

qsorry:	
  printmsg(sorrymsg);
  goto warmstart;

run_next_statement:
  while(*txtpos == ':')
    txtpos++;
  ignore_blanks();
  if(*txtpos == NL)
    goto execnextline;
  goto interperateAtTxtpos;

direct: 
  txtpos = program_end;
  if(*txtpos == NL)
    goto prompt;

interperateAtTxtpos:
  if(breakcheck())
  {
    printmsg(breakmsg);
    goto warmstart;
  }

  scantable(keywords);

#ifdef RASPI_PICO
  basic_cmd = table_index;
#endif

  switch(table_index)
  {
#ifndef RASPI_PICO
  case KW_DELAY:
    {
#ifdef ARDUINO
      expression_error = 0;
      val = expression();
      delay( val );
      goto execnextline;
#else
      goto unimplemented;
#endif
    }
#endif
  case KW_FILES:
    goto files;
  case KW_LIST:
    goto list;
#ifndef RASPI_PICO
  case KW_CHAIN:
    goto chain;
#endif
  case KW_LOAD:
    goto load;
  case KW_MEM:
    goto mem;
  case KW_NEW:
    if(txtpos[0] != NL)
      goto qwhat;
    program_end = program_start;
    init_var();
    goto prompt;
  case KW_RUN:
    current_line = program_start;
    goto execline;
  case KW_SAVE:
    goto save;
  case KW_NEXT:
    goto next;
  case KW_LET:
    goto assignment;
  case KW_IF:
    { long val;
    expression_error = 0;
    val = expression();
    if(expression_error || *txtpos == NL)
      goto qhow;
    if(val != 0)
      goto interperateAtTxtpos;
    goto execnextline;
    }
  case KW_GOTO:
    expression_error = 0;
    linenum = expression();
    if(expression_error || *txtpos != NL)
      goto qhow;
    {
     unsigned char *p = findNumberline();
     if ( p == NULL ) {
       printmsgNoNL("GOTO : ");
       printnum(linenum, -1);
       printmsgNoNL(" Missing. ");
       goto qhow;
     }
     current_line = p;
     goto execline;
    }

  case KW_GOSUB:
    goto gosub;
  case KW_RETURN:
    goto gosub_return; 
  case KW_REM:
  case KW_QUOTE:
    goto execnextline;	// Ignore line completely
  case KW_FOR:
    goto forloop; 
  case KW_INPUT:
    goto input; 
  case KW_PRINT:
  case KW_QMARK:
    goto print;
  case KW_POKE:
    goto poke;
  case KW_END:
  case KW_STOP:
    // This is the easy way to end - set the current line to the end of program attempt to run it
    if(txtpos[0] != NL)
      goto qwhat;
    current_line = program_end;
    goto execline;
  case KW_BYE:
    // Leave the basic interperater
    return;
#ifndef RASPI_PICO
  case KW_AWRITE:  // AWRITE <pin>, HIGH|LOW
    isDigital = false;
    goto awrite;
  case KW_DWRITE:  // DWRITE <pin>, HIGH|LOW
    isDigital = true;
    goto dwrite;
#endif
  case KW_RSEED:
    goto rseed;
#ifdef RASPI_PICO
  case KW_PINMODE:
    goto pinmode;
  case KW_SETRTC:
    goto setrtc;
  case KW_SLEEP:
    goto sleep;
  case KW_ADDA:
    goto adda;
  case KW_ALOAD:
    goto chain;
  case KW_FILECMD:
    goto file_cmd;
#endif
#ifdef ENABLE_TONES
  case KW_TONEW:
    alsoWait = true;
  case KW_TONE:
    goto tonegen;
  case KW_NOTONE:
    goto tonestop;
#endif

#ifdef ARDUINO
#ifdef ENABLE_EEPROM
  case KW_EFORMAT:
    goto eformat;
  case KW_ESAVE:
    goto esave;
  case KW_ELOAD:
    goto eload;
  case KW_ELIST:
    goto elist;
  case KW_ECHAIN:
    goto echain;
#endif
#endif

  case KW_DEFAULT:
    goto assignment;
  default:
    break;
  }

execnextline:
  if(current_line == NULL)		// Processing direct commands?
    goto prompt;
  current_line +=	 *((unsigned short*)(current_line+sizeof(LINENUM)));

execline:
  if(current_line == program_end) // Out of lines to run
    goto warmstart;
  txtpos = current_line+sizeof(LINENUM)+sizeof(unsigned short);
  goto interperateAtTxtpos;

#ifdef ARDUINO
#ifdef ENABLE_EEPROM
elist:
  {
    int i;
    for( i = 0 ; i < (E2END +1) ; i++ )
    {
      val = EEPROM.read( i );

      if( val == '\0' ) {
        goto execnextline;
      }

      if( ((val < ' ') || (val  > '~')) && (val != NL) && (val != CR))  {
        outchar( '?' );
      } 
      else {
        outchar( val );
      }
    }
  }
  goto execnextline;

eformat:
  {
    for( int i = 0 ; i < E2END ; i++ )
    {
      if( (i & 0x03f) == 0x20 ) outchar( '.' );
      EEPROM.write( i, 0 );
    }
    outchar( LF );
  }
  goto execnextline;

esave:
  {
    outStream = kStreamEEProm;
    eepos = 0;

    // copied from "List"
    list_line = findline();
    while(list_line != program_end) {
      printline();
    }
    outchar('\0');

    // go back to standard output, close the file
    outStream = kStreamSerial;
    
    goto warmstart;
  }
  
  
echain:
  runAfterLoad = true;

eload:
  // clear the program
  program_end = program_start;
  init_var();
  // load from a file into memory
  eepos = 0;
  inStream = kStreamEEProm;
  inhibitOutput = true;
  goto warmstart;
#endif /* ENABLE_EEPROM */
#endif

input:
  {
    unsigned char var;
    long value;
    ignore_blanks();
    if(*txtpos < 'A' || *txtpos > 'Z')
      goto qwhat;
    var = *txtpos;
    txtpos++;
    ignore_blanks();
    if(*txtpos != NL && *txtpos != ':')
      goto qwhat;
    tmptxtpos = txtpos;
inputagain:
    getln( '?' );
    toUppercaseBuffer();
    txtpos = program_end;
    ignore_blanks();
    expression_error = 0;
    value = expression();
    if(expression_error)
      goto inputagain;
    if( *txtpos != NL )
      goto inputagain;
    ((long*)variables_begin)[var-'A'] = value;
    txtpos = tmptxtpos;

    goto run_next_statement;
  }

forloop:
  {
    unsigned char var;
    long initial, step, terminal;
    ignore_blanks();
    if(*txtpos < 'A' || *txtpos > 'Z')
      goto qwhat;
    var = *txtpos;
    txtpos++;
    ignore_blanks();
    if(*txtpos != '=')
      goto qwhat;
    txtpos++;
    ignore_blanks();

    expression_error = 0;
    initial = expression();
    if(expression_error)
      goto qwhat;

    scantable(to_tab);
    if(table_index != 0)
      goto qwhat;

    terminal = expression();
    if(expression_error)
      goto qwhat;

    scantable(step_tab);
    if(table_index == 0)
    {
      step = expression();
      if(expression_error)
        goto qwhat;
    }
    else
      step = 1;
    ignore_blanks();
    if(*txtpos != NL && *txtpos != ':')
      goto qwhat;


    if(!expression_error && *txtpos == NL)
    {
      struct stack_for_frame *f;
      if(sp + sizeof(struct stack_for_frame) < stack_limit)
        goto qsorry;

      sp -= sizeof(struct stack_for_frame);
      f = (struct stack_for_frame *)sp;
      ((long *)variables_begin)[var-'A'] = initial;
      f->frame_type = STACK_FOR_FLAG;
      f->for_var = var;
      f->terminal = terminal;
      f->step     = step;
      f->txtpos   = txtpos;
      f->current_line = current_line;
      goto run_next_statement;
    }
  }
  goto qhow;

gosub:
  expression_error = 0;
  linenum = expression();
  if(!expression_error && *txtpos == NL)
  { unsigned char *p = findNumberline();
    if ( p == NULL ) {
       printmsgNoNL("GOSUB : ");
       printnum(linenum, -1);
       printmsgNoNL(" Missing. ");
       goto qhow;
    }
    struct stack_gosub_frame *f;
    if(sp + sizeof(struct stack_gosub_frame) < stack_limit)
      goto qsorry;

    sp -= sizeof(struct stack_gosub_frame);
    f = (struct stack_gosub_frame *)sp;
    f->frame_type = STACK_GOSUB_FLAG;
    f->txtpos = txtpos;
    f->current_line = current_line;
    current_line = p;
    goto execline;
  }
  goto qhow;

next:
  // Fnd the variable name
  ignore_blanks();
  if(*txtpos < 'A' || *txtpos > 'Z')
    goto qhow;
  next_var = txtpos;
  txtpos++;
  ignore_blanks();
  if(*txtpos != ':' && *txtpos != NL) {
    next_var = NULL;
    goto qwhat;
  }

gosub_return:
  // Now walk up the stack frames and find the frame we want, if present
  tempsp = sp;
  while(tempsp < program+sizeof(program)-1)
  {
    switch(tempsp[0])
    {
    case STACK_GOSUB_FLAG:
      next_var = NULL;
      if(table_index == KW_RETURN)
      {
        struct stack_gosub_frame *f = (struct stack_gosub_frame *)tempsp;
        current_line	= f->current_line;
        txtpos			= f->txtpos;
        sp += sizeof(struct stack_gosub_frame);
        goto run_next_statement;
      }
      // This is not the loop you are looking for... so Walk back up the stack
      tempsp += sizeof(struct stack_gosub_frame);
      break;
    case STACK_FOR_FLAG:
      // Flag, Var, Final, Step
      if(table_index == KW_NEXT)
      {
        struct stack_for_frame *f = (struct stack_for_frame *)tempsp;
        // Is the the variable we are looking for?
        if( *next_var == f->for_var)
        {
          long *varaddr = ((long *)variables_begin) + *next_var - 'A'; 
          next_var = NULL;
          *varaddr = *varaddr + f->step;
          // Use a different test depending on the sign of the step increment
          if((f->step > 0 && *varaddr <= f->terminal) || (f->step < 0 && *varaddr >= f->terminal))
          {
            // We have to loop so don't pop the stack
            txtpos = f->txtpos;
            current_line = f->current_line;
            goto run_next_statement;
          }
          // We've run to the end of the loop. drop out of the loop, popping the stack
          sp = tempsp + sizeof(struct stack_for_frame);
          goto run_next_statement;
        }
      }
      next_var = NULL;
      // This is not the loop you are looking for... so Walk back up the stack
      tempsp += sizeof(struct stack_for_frame);
      break;
    default:
      next_var = NULL;
      //printf("Stack is stuffed!\n");
      goto warmstart;
    }
  }
  // Didn't find the variable we've been looking for
  goto qhow;

assignment:
  {
    long value;
    long *var;

    if(*txtpos < 'A' || *txtpos > 'Z')
      goto qhow;
    var = (long *)variables_begin + *txtpos - 'A';
    txtpos++;

    ignore_blanks();

    if (*txtpos != '=')
      goto qwhat;
    txtpos++;
    ignore_blanks();
    expression_error = 0;
    value = expression();
    if(expression_error)
      goto qwhat;
    // Check that we are at the end of the statement
    if(*txtpos != NL && *txtpos != ':')
      goto qwhat;
    *var = value;
  }
  goto run_next_statement;
poke:
  {
    unsigned long value;
    unsigned char *address;
    boolean       l = false;

    // Work out where to put it
    expression_error = 0;
    value = (unsigned int)expression();
    if(expression_error)
      goto qwhat;
    address = (unsigned char *)value;

    // check for a comma
    ignore_blanks();
    if (*txtpos != ',')
      goto qwhat;
    txtpos++;
    ignore_blanks();

    // Now get the value to assign
    expression_error = 0;
    value = expression();
    if(expression_error)
      goto qwhat;
    //printf("Poke %p value %i\n",address, (unsigned char)value);
    // Check that we are at the end of the statement
    if(*txtpos == '#') {
      l = true;
      txtpos++;
    }
    if(*txtpos != NL && *txtpos != ':')
      goto qwhat;
#ifdef RASPI_PICO
    if( l && ((unsigned int)address & 0x0003) )
      goto qwhat;
    if( l )
      *((unsigned long*)address) = value;
    else
      *address = value;
#endif
  }
  goto run_next_statement;

list:
  linenum = testnum(); // Retuns 0 if no line found.
  if( linenum == 0xFFFF )
    goto qwhat;
  if( linenum == 0xFFFE )  linenum = 0;
  // Should be EOL
  if(txtpos[0] != NL)
    goto qwhat;

  // Find the line
  list_line = findline();
  while(list_line != program_end)
    printline();
  goto warmstart;

print:
  // If we have an empty list then just put out a NL
  if(*txtpos == ':' )
  {
    line_terminator();
    txtpos++;
    goto run_next_statement;
  }
  if(*txtpos == NL)
  {
    goto execnextline;
  }

  while(1)
  {
    ignore_blanks();
    if(print_quoted_string())
    {
      ;
    }
    else if(*txtpos == '"' || *txtpos == '\'')
      goto qwhat;
    else
    {
      char t = 0, d_min;
      long e;
      expression_error = 0;
      for ( d_min =0; *txtpos == '#' ; txtpos++ ) d_min++;
      if ( d_min > 0 ) {
        if ( *txtpos == '.' ) txtpos++;
        else goto qwhat;
      }
      if ( *txtpos == ESCHEX ) { t = 1; txtpos++; }
      e = expression();
      if(expression_error)
        goto qwhat;
      if ( t )  printhex(e, d_min);
      else      printnum(e, d_min);
    }

    // At this point we have three options, a comma or a new line
    if(*txtpos == ',')
      txtpos++;	// Skip the comma and move onto the next
    else if(txtpos[0] == ';' && (txtpos[1] == NL || txtpos[1] == ':'))
    {
      txtpos++; // This has to be the end of the print - no newline
      break;
    }
    else if(*txtpos == NL || *txtpos == ':')
    {
      line_terminator();	// The end of the print statement
      break;
    }
    else
      goto qwhat;	
  }
  goto run_next_statement;

mem:
  // memory free
  printmsgNoNL("Program Start: ");
  printhex((long)program_start, 8);
  printmsg("");
  printmsgNoNL("Start Free Memory: ");
  printhex((long)program_end, 8);
  printmsgNoNL(" - End Free Memory: ");
  printhex((long)(variables_begin - 1), 8);
  printmsg("");

  printnum(variables_begin-program_end, -1);
  printmsg(memorymsg);
#ifdef ARDUINO
#ifdef ENABLE_EEPROM
  {
    // eprom size
    printnum( E2END+1, -1 );
    printmsg( eeprommsg );
    
    // figure out the memory usage;
    val = ' ';
    int i;   
    for( i=0 ; (i<(E2END+1)) && (val != '\0') ; i++ ) {
      val = EEPROM.read( i );    
    }
    printnum( (E2END +1) - (i-1) , -1);
    
    printmsg( eepromamsg );
  }
#endif /* ENABLE_EEPROM */
#endif /* ARDUINO */
  goto run_next_statement;


  /*************************************************/

#if defined(ARDUINO) || defined(RASPI_PICO)
awrite: // AWRITE <pin>,val
dwrite:
  {
    short int pinNo;
    short int value;
    unsigned char *txtposBak;

    // Get the pin number
    expression_error = 0;
    pinNo = expression();
    if(expression_error)
      goto qwhat;

    // check for a comma
    ignore_blanks();
    if (*txtpos != ',')
      goto qwhat;
    txtpos++;
    ignore_blanks();


    txtposBak = txtpos; 
    scantable(highlow_tab);
    if(table_index != HIGHLOW_UNKNOWN)
    {
      if( table_index <= HIGHLOW_HIGH ) {
        value = 1;
      } 
      else {
        value = 0;
      }
    } 
    else {

      // and the value (numerical)
      expression_error = 0;
      value = expression();
      if(expression_error)
        goto qwhat;
    }
#ifdef RASPI_PICO
    if( isDigital ) {
      gpio_put( pinNo, value );
    } 
#else
    pinMode( pinNo, OUTPUT );
    if( isDigital ) {
      digitalWrite( pinNo, value );
    } 
    else {
      analogWrite( pinNo, value );
    }
#endif
  }
  goto run_next_statement;
#else
pinmode: // PINMODE <pin>, I/O
awrite: // AWRITE <pin>,val
dwrite:
  goto unimplemented;
#endif
#ifdef RASPI_PICO
pinmode: // PINMODE <pin>, I/O
  {
    short int pinNo;

    // Get the pin number
    expression_error = 0;
    pinNo = expression();
    if(expression_error)
      goto qwhat;

    // check for a comma
    ignore_blanks();
    if (*txtpos != ',')
      goto qwhat;
    txtpos++;
    ignore_blanks();

    scantable(inout_tab);
    if(table_index != INOUT_UNKNOWN)
    {
      if( table_index <= INOUT_OUT ) {
        gpio_init(pinNo);
        gpio_set_dir( pinNo, GPIO_OUT );
      } 
      else {
        gpio_init(pinNo);
        gpio_set_dir( pinNo, GPIO_IN );
      }
    } 
    else {
      goto qwhat;
    }
  }
  goto run_next_statement;

setrtc: // SETRTC "yyyy/mm/dd hh:mm:ss"
  {
    char ctime[22];
    ignore_blanks();
    if(!get_quoted_string(ctime,20))
      goto qwhat;
    else if(*txtpos == '"' || *txtpos == '\'')
      goto qwhat;
    else
    {
      if(set_rtc(ctime) )
        goto qwhat;
    }
  }

  goto run_next_statement;

sleep: // SLEEP 999 msec
  {
    long m_time;

    ignore_blanks();
    // Get time
    expression_error = 0;
    m_time = expression();
    if(expression_error)
      goto qwhat;

    sleep_ms((uint32_t)m_time);
  }

  goto run_next_statement;

adda: // ADDA Addr, add count (Add for Address) 
  {
    unsigned long *a_addr;
    long v_addr;
    
    ignore_blanks();
    if(*txtpos < 'A' || *txtpos > 'Z')
      goto qhow;
    a_addr = (unsigned long *)variables_begin + *txtpos - 'A';
    txtpos++;
    ignore_blanks();
    if (*txtpos != ',')
      goto qwhat;
    txtpos++;
    ignore_blanks();
    expression_error = 0;
    v_addr = expression();
    if(expression_error)
      goto qwhat;
    *a_addr += v_addr;
  }

  goto run_next_statement;

file_cmd:
  {
    int sz = file_operate(txtpos);
    if ( sz == 0 )
      goto qwhat;
    txtpos += sz;
  }
  goto run_next_statement;

#endif
  /*************************************************/
files:
  // display a listing of files on the device.
  // version 1: no support for subdirectories

#ifdef ENABLE_FILEIO
#ifdef RASPI_PICO
 {
    unsigned char dirname[130];

    // Work out the filename
    expression_error = 0;
    filenameWord(dirname);
    if(expression_error)
      goto qwhat;
    cmd_Files(dirname);
 }
#else
    cmd_Files();
#endif
  goto run_next_statement;
  // goto warmstart;
#else
  goto unimplemented;
#endif // ENABLE_FILEIO


chain:
  runAfterLoad = true;

load:

  // load from a file into memory
#ifdef ENABLE_FILEIO
  {
    unsigned char filename[130];
    FRESULT re;

    // Work out the filename
    expression_error = 0;
    filenameWord(filename);
    if(expression_error)
      goto qwhat;
#ifdef RASPI_PICO
    ignore_blanks();
    if( *txtpos != NL && *txtpos != 0 )
      goto qwhat;
    re = F_Open(4, filename, "r");
    if ( re ) {
      printmsgNoNL( "[Not Open:" );
      printmsgNoNL( filename );
      printmsgNoNL( "] " );
      goto qwhat;
    } 
    // clear the program
    if ( basic_cmd != KW_ALOAD ) {
      program_end = program_start;
      init_var();
    }
    inhibitOutput = true;

#endif

#ifdef ARDUINO
    // clear the program
    program_end = program_start;
    init_var();
    // Arduino specific
    if( !SD.exists( (char *)filename ))
    {
      printmsg( sdfilemsg );
    } 
    else {

      fp = SD.open( (const char *)filename );
      inStream = kStreamFile;
      inhibitOutput = true;
    }
#else // ARDUINO
    // Desktop specific
#endif // ARDUINO
    // this will kickstart a series of events to read in from the file.

  }
  goto warmstart;
#else // else ENABLE_FILEIO
  goto unimplemented;
#endif // ENABLE_FILEIO



save:
  // save from memory out to a file
#ifdef ENABLE_FILEIO
  {
    unsigned char filename[130];
    FRESULT re;

    // Work out the filename
    expression_error = 0;
    filenameWord(filename);
    if(expression_error)
      goto qwhat;
#ifdef RASPI_PICO
    linenum = testnum(); // Retuns 0 if no line found.
    if( linenum == 0xFFFF )
      goto qwhat;
    if( linenum == 0xFFFE )  linenum = 0;
    // Should be EOL
    if(txtpos[0] != NL)
      goto qwhat;

    list_line = findline();
    // copied from "List"
    if(list_line == program_end) {
      printmsg("No Program Statements!!");
    } else {

      re = F_Open(4, filename, "w");
      if ( re ) {
        printmsgNoNL( "[Not Open:" );
        printmsgNoNL( filename );
        printmsgNoNL( "] " );
        goto qwhat;
      } 

      while(list_line != program_end)
        printline();
      F_close(4);
    }
#endif
#ifdef ARDUINO
    // remove the old file if it exists
    if( SD.exists( (char *)filename )) {
      SD.remove( (char *)filename );
    }

    // open the file, switch over to file output
    fp = SD.open( (const char *)filename, FILE_WRITE );
    outStream = kStreamFile;

    // copied from "List"
    list_line = findline();
    while(list_line != program_end)
      printline();

    // go back to standard output, close the file
    outStream = kStreamSerial;

    fp.close();
#else // ARDUINO
    // desktop
#endif // ARDUINO
    goto warmstart;
  }
#else //  else ENABLE_FILEIO
  goto unimplemented;
#endif // ENABLE_FILEIO

rseed:
  {
    long value;

    //Get the pin number
    expression_error = 0;
    value = expression();
    if(expression_error)
      goto qwhat;

#ifdef ARDUINO
    randomSeed( value );
#else // ARDUINO
    srand( value );
#endif // ARDUINO
    goto run_next_statement;
  }

#ifdef ENABLE_TONES
tonestop:
  noTone( kPiezoPin );
  goto run_next_statement;

tonegen:
  {
    // TONE freq, duration
    // if either are 0, tones turned off
    short int freq;
    short int duration;

    //Get the frequency
    expression_error = 0;
    freq = expression();
    if(expression_error)
      goto qwhat;

    ignore_blanks();
    if (*txtpos != ',')
      goto qwhat;
    txtpos++;
    ignore_blanks();


    //Get the duration
    expression_error = 0;
    duration = expression();
    if(expression_error)
      goto qwhat;

    if( freq == 0 || duration == 0 )
      goto tonestop;

    tone( kPiezoPin, freq, duration );
    if( alsoWait ) {
      delay( duration );
      alsoWait = false;
    }
    goto run_next_statement;
  }
#endif /* ENABLE_TONES */
}

// returns 1 if the character is valid in a filename
static int isValidFnChar( unsigned char c )
{
  if( c >= '0' && c <= '9' ) return 1; // number
  if( c >= 'A' && c <= 'Z' ) return 1; // LETTER
  if( c >= 'a' && c <= 'z' ) return 1; // letter (for completeness)
  if( c == '_' ) return 1;
  if( c == '+' ) return 1;
  if( c == '.' ) return 1;
#ifdef RASPI_PICO
  if( c == '-' ) return 1;
  if( c == '/' ) return 1;
  if( c == ' ' ) return 1;
  if( c >= 0x80 && c < 0xff ) return 1;
#endif
  if( c == '~' ) return 1;  // Window~1.txt
  return 0;
}

void filenameWord(unsigned char *name)
{
  int s = 0;
  // SDL - I wasn't sure if this functionality existed above, so I figured i'd put it here
  expression_error = 0;

  // make sure there are no quotes or spaces, search for valid characters
  //while(*txtpos == SPACE || *txtpos == TAB || *txtpos == SQUOTE || *txtpos == DQUOTE ) txtpos++;
  if ( *txtpos == 0 || *txtpos == CR || *txtpos == NL ) {
    name[0] = 0;
    return;
  }
  if ( *txtpos != '"' ) {
    expression_error = 1;
    return;
  }
  txtpos++;
  if ( *txtpos == '"' ) {
    name[0] = 0;
    return;
  }
  while( s < 128 && isValidFnChar( *txtpos )) {
    name[s] = *txtpos;
    txtpos++; s++;
  }
  name[s] = 0;
  if( *txtpos != '"' ) {
    expression_error = 1;
    return;
  }
  txtpos++;
}

/***************************************************************************/
static void line_terminator(void)
{
  outchar(CR);
  outchar(NL);
}

/***********************************************************/
void setup()
{
printmsg(initmsg);
#ifdef RASPI_PICO
#ifdef ENABLE_FILEIO
  fatfs_init();
#endif
#endif
#ifdef ARDUINO
  Serial.begin(kConsoleBaud);	// opens serial port
  while( !Serial ); // for Leonardo
  
  Serial.println( sentinel );

#ifdef ENABLE_FILEIO
  initSD();

#ifdef ENABLE_AUTORUN
  if( SD.exists( kAutorunFilename )) {
    program_end = program_start;
    init_var();
    fp = SD.open( kAutorunFilename );
    inStream = kStreamFile;
    inhibitOutput = true;
    runAfterLoad = true;
  }
#endif /* ENABLE_AUTORUN */

#endif /* ENABLE_FILEIO */

#ifdef ENABLE_EEPROM
#ifdef ENABLE_EAUTORUN
  // read the first byte of the eeprom. if it's a number, assume it's a program we can load
  int val = EEPROM.read(0);
  if( val >= '0' && val <= '9' ) {
    program_end = program_start;
    init_var();
    inStream = kStreamEEProm;
    eepos = 0;
    inhibitOutput = true;
    runAfterLoad = true;
  }
#endif /* ENABLE_EAUTORUN */
#endif /* ENABLE_EEPROM */

#endif /* ARDUINO */
}


/***********************************************************/
static unsigned char breakcheck(void)
{
#ifdef ARDUINO
  if(Serial.available())
    return Serial.read() == CTRLC;
  return 0;
#else
#ifdef __CONIO__
  if(kbhit())
    return getch() == CTRLC;
  else
#endif
#ifdef RASPI_PICO
  return getchar_timeout_us(0) == CTRLC;
#endif
    return 0;
#endif
}
/***********************************************************/
static int inchar()
{
  int v;
#ifdef ENABLE_FILEIO
  if ( F_read(SAVE_FILENO, (unsigned char*)&v) ) {
    if ( inhibitOutput ) {
      F_close(SAVE_FILENO); // End of read file
#ifdef RASPI_PICO           
      inhibitOutput = false;
      if( runAfterLoad ) {  // For ALOAD
        runAfterLoad = false;
        triggerRun = true;
      } else {
        printmsgNoNL(okmsg);
      }
#else
      printmsgNoNL(okmsg);
#endif
      return NL;
    }
    // Not file load
  } else {    // Load file exec
    if ( v == NL ) v = CR;
    return v;
  }
#endif
#ifdef ARDUINO
  
  switch( inStream ) {
  case( kStreamFile ):
#ifdef ENABLE_FILEIO
    v = fp.read();
    if( v == NL ) v=CR; // file translate
    if( !fp.available() ) {
      fp.close();
      goto inchar_loadfinish;
    }
    return v;    
#else
#endif  // end ENABLE_FILEIO
     break;
  case( kStreamEEProm ):
#ifdef ENABLE_EEPROM
#ifdef ARDUINO
    v = EEPROM.read( eepos++ );
    if( v == '\0' ) {
      goto inchar_loadfinish;
    }
    return v;
#endif
#else
    inStream = kStreamSerial;
    return NL;
#endif  // end ENABLE_EEPROM
     break;
  case( kStreamSerial ):
  default:
    while(1)
    {
      if(Serial.available())
        return Serial.read();
    }
  }
  
inchar_loadfinish:
  inStream = kStreamSerial;
  inhibitOutput = false;

  if( runAfterLoad ) {
    runAfterLoad = false;
    triggerRun = true;
  }
  return NL; // trigger a prompt.
  
#else  // not ARDUINO
  // otherwise. desktop!
  int got = getchar();

  // translation for desktop systems
  if( got == LF ) got = CR;

  return got;
#endif  // end ARDUINO
}

/***********************************************************/
static void outchar(unsigned char c)
{
  if( inhibitOutput ) return;

#ifdef ARDUINO
  #ifdef ENABLE_FILEIO
    if( outStream == kStreamFile ) {
      // output to a file
      fp.write( c );
    } 
    else
  #endif
  #ifdef ARDUINO
  #ifdef ENABLE_EEPROM
    if( outStream == kStreamEEProm ) {
      EEPROM.write( eepos++, c );
    }
    else 
  #endif /* ENABLE_EEPROM */
  #endif /* ARDUINO */
    Serial.write(c);

#else
#ifdef RASPI_PICO
  if( !F_write(SAVE_FILENO, c) ) return;
#endif
  putchar(c);
#endif
}

/***********************************************************/
/* SD Card helpers */

#if ARDUINO && ENABLE_FILEIO

static int initSD( void )
{
  // if the card is already initialized, we just go with it.
  // there is no support (yet?) for hot-swap of SD Cards. if you need to 
  // swap, pop the card, reset the arduino.)

  if( sd_is_initialized == true ) return kSD_OK;

  // due to the way the SD Library works, pin 10 always needs to be 
  // an output, even when your shield uses another line for CS
  pinMode(10, OUTPUT); // change this to 53 on a mega

  if( !SD.begin( kSD_CS )) {
    // failed
    printmsg( sderrormsg );
    return kSD_Fail;
  }
  // success - quietly return 0
  sd_is_initialized = true;

  // and our file redirection flags
  outStream = kStreamSerial;
  inStream = kStreamSerial;
  inhibitOutput = false;

  return kSD_OK;
}
#endif

#if ENABLE_FILEIO
#ifdef RASPI_PICO
void cmd_Files(char *str_dir)
{
    FRESULT re;
    DIR Dir;				/* Directory object */
    UINT acc_files, acc_dirs;
    DWORD dw, acc_size;
    FILINFO Finfo;
    FATFS *fs;
    char buf[40];

    if( str_dir == NULL )  str_dir = "";
    printmsg("===== File List =====");
    if( !fatfs_is_ready() ) {
      printmsg("SD Card is Empty !!");
      return;
    } 
    re = f_opendir(&Dir, str_dir);
    if (re) {
        printmsg("Dir Not Found");
        return;
    }
    acc_size = acc_dirs = acc_files = 0;
    for(;;) {
        re = f_readdir(&Dir, &Finfo);
        if ((re != FR_OK) || !Finfo.fname[0] ) break;
        if (Finfo.fattrib & AM_DIR) {
	    acc_dirs++;
        } else {
            acc_files++; acc_size += Finfo.fsize;
        }
        sprintf(buf, "%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu%2s ",
            (Finfo.fattrib & AM_DIR) ? 'D' : '-',
            (Finfo.fattrib & AM_RDO) ? 'R' : '-',
            (Finfo.fattrib & AM_HID) ? 'H' : '-',
            (Finfo.fattrib & AM_SYS) ? 'S' : '-',
            (Finfo.fattrib & AM_ARC) ? 'A' : '-',
            (Finfo.fdate >> 9) + 1980, (Finfo.fdate >> 5) & 15, Finfo.fdate & 31,
            (Finfo.ftime >> 11), (Finfo.ftime >> 5) & 63,
            Finfo.fsize > 1024 ? (Finfo.fsize+511)/1024 :Finfo.fsize, Finfo.fsize > 1024 ? "KB" : "B");
        printmsgNoNL(buf);
        printmsg(Finfo.fname);
    }
    sprintf( buf,"%4u File(s),%10lu KB ",acc_files, (acc_size+1023)/1024);
    printmsgNoNL(buf);
    sprintf( buf,"%4u Dir(s)", acc_dirs);
    printmsgNoNL(buf);
    re = f_getfree("", &dw, &fs);
    if (re == FR_OK) {
       sprintf( buf,", %10lu KB free", ((dw * fs->csize * 512)+1023)/1024);
       printmsg(buf);
    } else {
       sprintf( buf, " Err:%X\n", re);
       printmsg(buf);
    }
}
#else
void cmd_Files( void )
{
  File dir = SD.open( "/" );
  dir.seek(0);

  while( true ) {
    File entry = dir.openNextFile();
    if( !entry ) {
      entry.close();
      break;
    }

    // common header
    printmsgNoNL( indentmsg );
    printmsgNoNL( (const unsigned char *)entry.name() );
    if( entry.isDirectory() ) {
      printmsgNoNL( slashmsg );
    }

    if( entry.isDirectory() ) {
      // directory ending
      for( int i=strlen( entry.name()) ; i<16 ; i++ ) {
        printmsgNoNL( spacemsg );
      }
      printmsgNoNL( dirextmsg );
    }
    else {
      // file ending
      for( int i=strlen( entry.name()) ; i<17 ; i++ ) {
        printmsgNoNL( spacemsg );
      }
      printUnum( entry.size() );
    }
    line_terminator();
    entry.close();
  }
  dir.close();
}
#endif
#endif
