/*----------------------------------------------------------------------*/
/* Ev.Board: Raspberry pi PICO                                          */
/* Console: N81 115200bps                                               */
/* MMC/SDC: SPI mode                      K.S. (c)2021                  */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include "pico/stdlib.h"
#include "hardware/spi.h"
#include "hardware/irq.h"
#include "hardware/pwm.h"
// #include "hardware/uart.h"
#include "hardware/rtc.h"
#include "pico/util/datetime.h"

/* Define configrations for application */
#define SOUND_DEFINED 0
//#define AKDK_SOCKET             // If In SD card is Low Active, then is defined.
/* ------------------------------------ */
#ifdef DDEBUG
#define DBG_NOP()	asm volatile("nop\n")
#else
#define DBG_NOP()
#endif
// define LED
#define LED_PIN 25
#define LED_INIT()      gpio_init(LED_PIN);gpio_set_dir(LED_PIN,GPIO_OUT);gpio_put(LED_PIN, 0)
#define LED_ON()	pwm_set_gpio_level(PICO_DEFAULT_LED_PIN, (FPWM+1)/4)
#define LED_OFF()	pwm_set_gpio_level(PICO_DEFAULT_LED_PIN, (FPWM+1)/100)
#define SIO_GPIO_OE(x)		*((unsigned long*)0xD0000020)=1<<(x)
#define SIO_GPIO_OUT_XOR(x)	*((unsigned long*)0xD000001C)=1<<(x)
#define GPIO_CTL25(x)		*((unsigned long*)0x400140CC)=(x)
#define PWM4_CNTL(x)		*((unsigned long*)0x40050050)=(x)

#define SET_HARD_FAULT_HANDLER() *((unsigned long*)0x2000000C)=(unsigned long)Ev_Board_Hardfault_Handler

#define SPI_PORT spi0
#define MISO_PIN 16
#define CS_PIN   17
#define SCK_PIN  18
#define MOSI_PIN 19
#define MMC_CD_PIN 22
#define MMC_CD_INIT()   gpio_init(MMC_CD_PIN);gpio_set_dir(MMC_CD_PIN,GPIO_IN);gpio_pull_up(MMC_CD_PIN)
//      MMC_CD      On = Card set , Off= Card set <<==== Please SELECT
#ifdef AKDK_SOCKET
#define MMC_CD		!gpio_get(MMC_CD_PIN)
#else
#define MMC_CD		gpio_get(MMC_CD_PIN)
#endif

#define	CS_HIGH()		CS_h()	/* Set MMC CS "high" */
#define CS_LOW()		CS_l()	/* Set MMC CS "low" */
static void CS_h();
static void CS_l();

#define	INIT_PORT()	spi_port_ini()	/* Initialize MMC control port (CS/CLK/DI:output, DO:input) */
#define DLY_US(n)	sleep_us(n)	/* Delay n microseconds */
BYTE xchg_spi (BYTE dat);	/* Data to send */

void spi_port_ini(void);
void rcvr_spi_multi (BYTE *buff,	UINT btr);
BYTE xchg_spi (BYTE dat);
void rcvr_spi_multi (BYTE *buff, UINT btr);

#if FF_FS_READONLY == 0
/* Send multiple byte */
void xmit_spi_multi (const BYTE *buff,UINT btx);			/* Number of bytes to send (512) */
#endif

/* TIMER */
#define EVBOARD_CLEAR_INTFLAG()    pwm_clear_irq(pwm_gpio_to_slice_num(PICO_DEFAULT_LED_PIN))
void Ev_Board_Timer_init(irq_handler_t handler);
#define FPWM	1249
/* TIMER end */

/* UART */
#define UART_ID uart0
#define BAUD_RATE 115200

#define UART_TX_PIN 0
#define UART_RX_PIN 1

void Ev_Board_uart_init(void);

#define x_putc(c)    uart_putc(UART_ID,(c))
#define xgetc()    uart_getc(UART_ID)

/* UART end */

struct file_obj {
	FIL     fil;
	char	opn_f;
	char	end_f;
};
#define SAVE_FILENO 4
#define FILENO_CHECK(x) ((x) < 0 || (x) > 4)
#define USERNO_CHECK(x) ((x) < 0 || (x) > 3)

void fatfs_init(void);
FRESULT F_Open (
	int   file_no,		/* file object number */
	const TCHAR* path,	/* Pointer to the file name */
	const TCHAR* mode	/* Access mode and file open mode flags */
);
int F_write(int fn, unsigned char c);
int F_writeN(int fn, unsigned char *buf, UINT sz);
int F_read(int fn, unsigned char *c);
int F_gets( unsigned char *buf,int siz, int fn);
int F_puts(int fn, unsigned char *buf);

void F_close(int fn);
int set_rtc(char *ctime);
DWORD get_rtc_date(void);
DWORD get_rtc_time(void);
DWORD get_rtime(void);

int file_operate(unsigned char *cmd);
void fatfs_Not_ready(void);
int fatfs_is_ready(void);
void Ev_Board_Hardfault_Handler(void);
void Disk_Timer_Handler (void);
